class BaseTranslatePrompter:
    def __str__(self):
        return '''You are an advanced {src_lang} to {target_lang} translator.
'The input format: {{"total_number": <total-number>, "list": ["1-<sentence-1>", "2-<sentence-2>", "3-<sentence-3>", ...]}}.
'The output format: {{"total_number": <total-translated-number>, "list": ["1-<translated-sentence-1>", "2-<translated-sentence-2>", "3-<translated-sentence-3>", ..., "<total-translated-number>-<last-translated-sentence>"]}}.
'Please remember to add an order number before each translated sentence, as the output format.
'Return only the content inside the two curly brackets and nothing else.
'Do not merge or eliminate any sentence.
'Keep the number of the translated sentence in the returned list identical to that of the original list.
'Ensure that the correspondence between the original and translated list is maintained.
'DO NOT break the output format.
'Utilize context to rectify any inappropriate words.
'If a sentence is not suitable for translation, please do not translate it, but you need to complete the translation of other suitable sentences.
'If you are not sure whether a sentence is suitable for translation, please translate it.
'Even if encountering sentences that are not suitable for translation, please maintain the output format.'
'Even if only one sentence is translated, please maintain the output format. Dont add any other words.
'Use '<' and '>' to replace the double quote in the translated sentences.'''


class LovelyPrompter(BaseTranslatePrompter):
    def __str__(self):
        return super().__str__() + 'Use lovely colloquial expressions to translate each sentence.'


prompter_map = {
    'base_trans': BaseTranslatePrompter,
    'lovely_trans': LovelyPrompter
}


def format_texts(texts):
    return f'{{"total_number": {len(texts)}, "list": {list2str(texts)}}}'


def list2str(texts):
    """
    To fit the prompter format, add order number to each sentence and use double quota to wrap each element in the list.
    """

    text = [f'"{i}-{str(text)}"' for i, text in enumerate(texts, start=1)]
    return f"[{', '.join(text)}]"



import os

from faster_whisper import WhisperModel
from tqdm import tqdm

from openlrc.logger import logger
from openlrc.lrc import LRC, LRCOptimizer
from openlrc.prompter import prompter_map
from openlrc.utils import Timer, change_ext, extend_filename


class LRCer:
    def __init__(self, model_name='large-v2'):
        self.model = WhisperModel(model_name, compute_type="float16")

    def __call__(self, audio_path, target_lang='zh-cn', prompter='base_trans'):
        if prompter not in prompter_map:
            raise ValueError(f'Prompter {prompter} not found.')
        prompter = prompter_map[prompter]()  # Initialize appropriate prompter

        transcribed_lrc_path = change_ext(extend_filename(audio_path, '_transcribed'), 'lrc')
        if not os.path.exists(transcribed_lrc_path):
            logger.info(f'Not found transcribed lrc file: {transcribed_lrc_path}')
            with Timer('Transcribing...'):
                # Temporarily workaround for https://github.com/guillaumekln/faster-whisper/issues/71
                # TODO: Should remove `temperature=0` to enhance whisper performance after issue fixed
                segments, info = self.model.transcribe(audio_path, word_timestamps=False, temperature=0)

                logger.info(f'Length of audio {audio_path}: {LRC.format_timestamp(info.duration)}')
                logger.info(f'Detected language: {info.language}, with probability: {info.language_probability}')

                # From generator to list with progress bar shown
                seg_list = []
                for segment in tqdm(segments, 'Transcribing'):
                    seg_list.append(segment)
                logger.debug(f'Transcribed fast-whisper Segments: {seg_list}')

            # Save the transcribed lrc
            self.to_lrc(seg_list, name=transcribed_lrc_path, lang=info.language)  # xxx_transcribed.lrc
        else:
            logger.info(f'Found transcribed lrc file: {transcribed_lrc_path}')

        transcribed_optimized_path = self.post_process(transcribed_lrc_path)  # xxx_transcribed_optimized.lrc

        # Translate the transcribed lrc
        transcribed_optimized_lrc = LRC(transcribed_optimized_path)

        with Timer('Translating...'):
            lrc_name = transcribed_optimized_lrc.translate(
                prompter=prompter, target_lang=target_lang
            )  # xxx_transcribed_optimized_translated.lrc

        self.post_process(lrc_name, output_lrc_name=change_ext(audio_path, 'lrc'),
                          remove_files=[
                              transcribed_optimized_path,  # xxx_transcribed_optimized.lrc
                              lrc_name  # xxx_transcribed_optimized_translated.lrc
                          ])  # xxx.lrc

    @staticmethod
    def to_lrc(segments, name, lang):
        """
        Convert the segments into lrc format.
        """
        with open(name, 'w', encoding='utf-8') as f:
            print(f'LRC generated by https://github.com/zh-plus/Open-Lyrics, lang={lang}', file=f, flush=True)
            for i, segment in enumerate(segments):
                print(
                    f'[{LRC.format_timestamp(segment.start)}] {segment.text}',
                    file=f,
                    flush=True,
                )

                if i != len(segments) - 1 and segment.end != segments[i + 1].start:
                    print(f'[{LRC.format_timestamp(segment.end)}]', file=f, flush=True)

        logger.info(f'File saved to {name}')

    @staticmethod
    def post_process(lrc_name, output_lrc_name=None, remove_files=None):
        lrc_optimizer = LRCOptimizer(lrc_name)
        lrc_optimizer.perform_all()
        optimized_name = lrc_optimizer.save(output_lrc_name=output_lrc_name)

        # Remove intermediate files
        if remove_files:
            for file in remove_files:
                os.remove(file)

        return optimized_name



import json
import time
from os.path import splitext

import tiktoken

from openlrc.logger import logger


def json2dict(json_str):
    """ Convert json string to python dict. """

    try:
        result = json.loads(json_str)
        return result
    except json.decoder.JSONDecodeError as e:
        logger.warning(f'Fail to convert into json: \n {json_str}\n Trying to fix...')

    # Try to fix the json string, only keep the content from first '{' to last '}'
    fixed_json_str1 = json_str[json_str.find('{'):json_str.rfind('}') + 1]
    logger.warning(
        f'Trying to fix the json string by keep only "{{content}}": \n {json_str}\n Into: \n {fixed_json_str1}\n')
    try:
        result = json.loads(fixed_json_str1)
        return result
    except json.decoder.JSONDecodeError:
        logger.warning(f'Failed to convert into json: \n {fixed_json_str1}\n Trying to fix...')

    # Try to replace chinese "，" with english ","
    fixed_json_str2 = fixed_json_str1.replace('，', ',')
    logger.warning(
        f'Trying to fix the json string by replace chinese quote with eng quote: \n {fixed_json_str1}\n Into: \n {fixed_json_str2}\n'
    )
    try:
        result = json.loads(fixed_json_str2)
        return result
    except json.decoder.JSONDecodeError:
        logger.warning(f'Failed to convert into json: \n {fixed_json_str2}\n Trying to fix...')

    # The content after last found " should be "}]"
    fixed_json_str3 = fixed_json_str2[:fixed_json_str2.rfind('"') + 1] + ']}'
    try:
        result = json.loads(fixed_json_str3)
        return result
    except json.decoder.JSONDecodeError:
        logger.warning(f'Failed to convert into json: \n {fixed_json_str3}\n Trying to fix...')

    # Try to ensure the sentence lists between "[]" do not contain extra "
    start_idx = fixed_json_str3.find('[')
    sentences = fixed_json_str3[start_idx + 2:-3]  # also remove the first and last "
    sentences = sentences.split('","')
    sentences = [sentence.replace('"', '\\"') for sentence in sentences]
    sentences = '","'.join(sentences)
    fixed_json_str4 = fixed_json_str3[:start_idx + 2] + sentences + fixed_json_str3[-3:]
    try:
        result = json.loads(fixed_json_str4)
        return result
    except json.decoder.JSONDecodeError as e:
        # Save the json string to file
        with open('test_return.json', 'w', encoding='utf-8') as f:
            f.write(fixed_json_str4)

        logger.info(f'The json file is saved to test_return.json')

        logger.error(f'Failed to convert returned content into json: \n {fixed_json_str4}\n\n Fix failed!')

        raise e


def get_token_number(text, model="gpt-3.5-turbo"):
    encoder = tiktoken.encoding_for_model(model)

    return len(encoder.encode(text))


def change_ext(filename, ext):
    return f'{splitext(filename)[0]}.{ext}'


def extend_filename(filename, extend):
    name, ext = splitext(filename)
    return f'{name}{extend}{ext}'


class Timer:
    def __init__(self, task=""):
        self._start = None
        self._stop = None
        self.task = task

    def start(self):
        if self.task:
            logger.info(f'Start {self.task}')
        self._start = time.perf_counter()

    def stop(self):
        self._stop = time.perf_counter()
        logger.info(f'{self.task} Elapsed: {self.elapsed:.2f}s')

    @property
    def elapsed(self):
        if self._start is None:
            raise RuntimeError("Timer not started")
        if self._stop is None:
            raise RuntimeError("Timer not stopped")
        return self._stop - self._start

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()



import math
import sys
from dataclasses import dataclass
from typing import List, Union

from langcodes import Language

from openlrc.chatbot import GPTBot
from openlrc.exceptions import SameLanguageException
from openlrc.logger import logger
from openlrc.prompter import BaseTranslatePrompter, format_texts
from openlrc.utils import extend_filename, json2dict


@dataclass
class LRCElement:
    """
    Save a LRC format element.
    """
    start: float
    end: Union[float, None]
    text: str

    @property
    def duration(self):
        if self.end:
            return self.end - self.start
        else:
            return sys.maxsize  # Fake int infinity


class LRC:
    """
    Save a LRC format data.
    """

    def __init__(self, lrc_name=None):
        self.lrc_name = lrc_name
        self.lang = self.get_lang(lrc_name)
        self.elements: List[LRCElement] = []

        if lrc_name:
            self.elements = self.read_lrc(file=lrc_name)

    def read_lrc(self, file):
        """
        Read lrc file and return a list of LRCElement.
        """
        with open(file, 'r', encoding='utf-8') as f:
            lines = f.readlines()

        elements = []

        for line in lines:
            if line.startswith('['):
                start, text = line.split(']', 1)
                start = self.parse_timestamp(start[1:])
                text = text.strip()
                elements.append(LRCElement(start, None, text))

        for i, element in enumerate(elements[:-1]):
            element.end = elements[i + 1].start

        return elements

    def save_lrc(self, file):
        """
        Save lrc file.
        """
        with open(file, 'w', encoding='utf-8') as f:
            print(f'LRC generated by https://github.com/zh-plus/Open-Lyrics, lang={self.lang}', file=f, flush=True)
            for i, element in enumerate(self.elements):
                print(
                    f'[{self.format_timestamp(element.start)}] {element.text}',
                    file=f,
                    flush=True,
                )

                if i != len(self.elements) - 1 and element.end != self.elements[i + 1].start:
                    print(f'[{self.format_timestamp(element.end)}]', file=f, flush=True)

        return file

    def get_system_prompt(self, src_lang, target_lang, prompter):
        system_prompt = str(prompter).format(
            src_lang=Language.get(self.lang if not src_lang else src_lang).display_name('en'),
            target_lang=Language.get(target_lang).display_name('en')
        )
        # Prevent translating text into Traditional Chinese
        if target_lang == 'zh-cn':
            system_prompt.replace(Language.get(target_lang).display_name('en'), 'Mandarin Chinese')

        return system_prompt

    def translate(self, prompter=BaseTranslatePrompter(), chunk_size=20, src_lang=None, target_lang='zh-cn',
                  intercept_line=None, force_translate=False):
        """
        Use GPT-3.5 to translate lyrics.
        TODO: dynamically adjust the chunk size.
        :param prompter: Translate prompter.
        :param chunk_size: Use smaller chunk size to avoid exceeding the token limit & output complete message.
        :param src_lang: Source language.
        :param target_lang: Target language.
        :param intercept_line: Intercepted lyrics line number.
        :param force_translate: Force translation even if the source language is the same as the target language.
        :return: The translated lrc file path.
        """
        if not force_translate and \
                Language.get(self.lang if not src_lang else src_lang).language_name() \
                == Language.get(target_lang).language_name():
            raise SameLanguageException()

        # Resulted json content
        json_content = {'total_number': 0, 'list': []}

        system_prompt = self.get_system_prompt(src_lang, target_lang, prompter)
        translate_bot = GPTBot(system_prompt=system_prompt)

        lyrics = [element.text for element in self.elements[:intercept_line]]
        # Split lyrics into different chunks
        chunks = [lyrics[i:i + chunk_size] for i in range(0, len(lyrics), chunk_size)]
        user_prompts = [format_texts(chunk) for chunk in chunks]  # Format the chunks into a single string

        logger.info(f'Translating {len(user_prompts)} user_prompts of lyrics with async call.')

        # Start translate
        responses = translate_bot.message(user_prompts)
        for i, response in enumerate(responses):
            content = response.choices[0].message.content
            logger.debug(f'Target content - chunk{i}: {content}')

            chunk_json_content = json2dict(content)
            logger.debug(f'Length of the translated chunk: {len(chunk_json_content["list"])}')

            chunk_size = len(chunks[i])
            # Helping OpenAI clean up their mess.
            if len(chunk_json_content['list']) < chunk_size:
                logger.warning('The number of translated sentences is less than that of the original list. '
                               'Add <MANUALLY-ADDED> label')
                chunk_json_content['list'] += ['<MANUALLY-ADDED>'] * (chunk_size - len(chunk_json_content['list']))
            elif len(chunk_json_content['list']) > chunk_size:
                logger.warning('The number of translated sentences is more than that of the original list. Truncated')
                chunk_json_content['list'] = chunk_json_content['list'][:chunk_size]

            json_content['total_number'] += chunk_json_content['total_number']
            json_content['list'] += chunk_json_content['list']

        # Remove the order number at the front of each sentence
        for i, text in enumerate(json_content['list']):
            json_content['list'][i] = text[text.find('-') + 1:]

        # Replace the original lyrics with the translated lyrics
        for i in range(len(self.elements)):
            self.elements[i].text = json_content['list'][i]
        self.lang = target_lang

        return self.save_lrc(extend_filename(self.lrc_name, '_translated'))

    @staticmethod
    def format_timestamp(seconds: float):
        assert seconds >= 0, "non-negative timestamp expected"
        milliseconds = round(seconds * 1000.0)

        minutes = milliseconds // 60_000
        milliseconds -= minutes * 60_000

        seconds = milliseconds // 1_000
        milliseconds -= seconds * 1_000

        return f"{minutes:02d}:{seconds:02d}.{milliseconds:02d}"

    @staticmethod
    def parse_timestamp(time_stamp):
        minutes, seconds = time_stamp.split(':')
        seconds, milliseconds = seconds.split('.')
        return int(minutes) * 60 + int(seconds) + int(milliseconds) / 1000.0

    @staticmethod
    def get_lang(lrc_name):
        with open(lrc_name, 'r', encoding='utf-8') as f:
            first_line = f.readline()

        if first_line.startswith('LRC generated by'):
            lang = first_line.split('lang=')[-1].strip()
            return lang


class LRCOptimizer:
    def __init__(self, lrc: Union[str, LRC]):
        if isinstance(lrc, str):
            lrc = LRC(lrc)

        self.lrc = lrc

    @property
    def lrc_name(self):
        return self.lrc.lrc_name

    def merge_same_lyrics(self):
        """
        Merge the same text.
        """
        new_elements = []

        for i, element in enumerate(self.lrc.elements):
            if i == 0 or element.text != new_elements[-1].text:
                new_elements.append(element)
            else:
                new_elements[-1].end = element.end

        logger.debug(f'Merge same text: {len(self.lrc.elements)} -> {len(new_elements)}')

        self.lrc.elements = new_elements

    def merge_short_lyrics(self, threshold=2):
        """
        Merge the short text.
        """
        new_elements = []

        for i, element, in enumerate(self.lrc.elements):
            if i == 0 or element.duration >= threshold:
                new_elements.append(element)
            else:
                new_elements[-1].text += ' ' + element.text
                new_elements[-1].end = element.end

        logger.debug(f'Merge short text: {len(self.lrc.elements)} -> {len(new_elements)}')

        self.lrc.elements = new_elements

    def merge_same_words(self):
        """
        Merge the same pattern in one lyric.
        :return:
        """
        elements = self.lrc.elements

        def get_repeat(text):
            """
            Check if the text is repeated for [1-4] words.
            """
            for i in range(1, 5):
                repeating_num = math.floor(len(text) / float(i))
                if text[:i] * repeating_num == text[:i * repeating_num]:
                    return text[:i]
            return None

        for i in range(len(elements)):
            repeat_text = get_repeat(elements[i].text)
            if repeat_text:
                elements[i].text = repeat_text + '...(Repeat)'
                logger.debug(f'Merge same words: {repeat_text}')

        logger.debug('Merge same words done.')

        self.lrc.elements = elements

    def cut_long_lyrics(self, threshold=125, keep=20):
        elements = self.lrc.elements

        for i, element in enumerate(elements):
            if len(element.text) > threshold:
                logger.warning(f'Cut long text: {element.text}\nInto: {element.text[:keep]}...')
                elements[i].text = element.text[:keep] + f'(Cut to {keep})'

        logger.debug('Cut long text done.')

        self.lrc.elements = elements

    def perform_all(self):
        for _ in range(2):
            self.merge_same_lyrics()
            self.merge_short_lyrics()
            self.merge_same_words()
            self.cut_long_lyrics()

    def save(self, output_lrc_name=None):
        optimized_name = extend_filename(self.lrc_name, '_optimized') if not output_lrc_name else output_lrc_name
        self.lrc.save_lrc(optimized_name)
        logger.info(f'Optimized LRC file saved to {optimized_name}')
        return optimized_name



class SameLanguageException(Exception):
    """
    Raised when the source language and target language are the same.
    """

    def __init__(self):
        super().__init__(
            'Source language and target language are the same, no need to translate. '
            'If you want to translate, set force_translate=True.'
        )


class ChatBotException(Exception):
    """
    Raised when chatbot fails to generate response.
    """

    def __init__(self, message):
        super().__init__(message)



import asyncio
import os
import time
from typing import List, Union

import openai
from aiohttp import ClientSession

from openlrc.exceptions import ChatBotException
from openlrc.logger import logger
from openlrc.utils import get_token_number


class GPTBot:
    def __init__(self, system_prompt="", model='gpt-3.5-turbo', retry=5, max_async=16, fee_limit=0.1):
        openai.api_key = os.getenv("OPENAI_API_KEY")

        self.model = model
        self.system_prompt = system_prompt
        self.retry = retry
        self.max_async = max_async
        self.fee_limit = fee_limit

    def get_fee(self, user_prompts):
        """
        Calculate the total fee for the given user prompts.
        Pricing info from https://openai.com/pricing
        """
        # Pricing for 1k token
        pricing = {
            'gpt-3.5': (0.002, 0.002),
            'gpt-4': (0.03, 0.06),
            'gpt-4-32k': (0.06, 0.12)
        }

        system_token, user_token = get_token_number(self.system_prompt), get_token_number(' '.join(user_prompts))

        if self.model.startswith('gpt-3.5'):
            prompt_price, completion_price = pricing['gpt-3.5']
        elif self.model.startswith('gpt-4-32k'):
            prompt_price, completion_price = pricing['gpt-4-32k']
        elif self.model.startswith('gpt-4'):
            prompt_price, completion_price = pricing['gpt-4']
        else:
            raise ChatBotException('Fail to get fee. Invalid model name.')

        total_price = ((system_token + user_token) * prompt_price + user_token * completion_price) / 1000

        return total_price

    async def _create_achat(self, user_prompt):
        logger.debug(f'Raw content: {user_prompt}')

        response = None
        for _ in range(self.retry):
            try:
                response = openai.ChatCompletion.create(
                    model="gpt-3.5-turbo",
                    messages=[
                        {
                            'role': 'system',
                            'content': self.system_prompt
                        },
                        {
                            'role': 'user',
                            'content': user_prompt
                        }
                    ]
                )
                break
            except openai.error.RateLimitError:
                logger.warning('Rate limit exceeded. Wait 10s before retry.')
                time.sleep(10)
            except openai.error.APIConnectionError:
                logger.warning('API connection error. Wait 30s before retry.')
                time.sleep(30)

        if not response:
            raise ChatBotException('Failed to create a chat.')

        return response

    async def _amessage(self, messages):
        """
        Async send messages to the GPT chatbot.
        """
        async with openai.aiosession.get(ClientSession()):
            results = await asyncio.gather(*(self._create_achat(message) for message in messages))

            return results

    def message(self, messages: Union[str, List]):
        """
        Send messages to the GPT chatbot.
        """
        if isinstance(messages, str):
            messages = [messages]

        # Calculate the total sending token number and approximated billing fee.
        token_numbers = [get_token_number(message + self.system_prompt) for message in messages]

        # if any of token number exceeds the limit, raise an exception.
        if any(token_number > 2048 for token_number in token_numbers):
            raise ChatBotException(f'Token number {max(token_numbers)} exceeds the limit.')

        # if the approximated billing fee exceeds the limit, raise an exception.
        approximated_fee = self.get_fee(messages)
        logger.info(f'Approximated billing fee: {approximated_fee} US$')
        if approximated_fee > self.fee_limit:
            raise ChatBotException(f'Approximated billing fee {approximated_fee} '
                                   f'exceeds the limit: {self.fee_limit}$.')

        try:
            results = asyncio.run(self._amessage(messages))
        except ChatBotException as e:
            raise e

        return results
